import numpy as np
from .loader.base_loader import BaseDataLoader

class DataApi:
    def __init__(self, data_loader: BaseDataLoader, default_stock_id: str, defaut_freq: str):
        self._data_loader = data_loader
        self._defaut_freq = defaut_freq
        self._default_stock_id = default_stock_id

    def close(self, code=None, freq=None):
        if code is None:
            code = self._default_stock_id
        if freq is None:
            freq = self._defaut_freq
        df = self._data_loader.get_stock_data_by_date(code, freq)
        if df.empty:
            return np.array([])
        return np.array(df['Close'], dtype=float)
    
    def high(self, code=None, freq=None):
        if code is None:
            code = self._default_stock_id
        if freq is None:
            freq = self._defaut_freq
        df = self._data_loader.get_stock_data_by_date(code, freq)
        if df.empty:
            return np.array([])
        return np.array(df['High'], dtype=float)
    
    def low(self, code=None, freq=None):
        if code is None:
            code = self._default_stock_id
        if freq is None:
            freq = self._defaut_freq
        df = self._data_loader.get_stock_data_by_date(code, freq)
        if df.empty:
            return np.array([])
        return np.array(df['Low'], dtype=float)
    
    def open(self, code=None, freq=None):   
        if code is None:
            code = self._default_stock_id
        if freq is None:
            freq = self._defaut_freq
        df = self._data_loader.get_stock_data_by_date(code, freq)
        if df.empty:
            return np.array([])
        return np.array(df['Open'], dtype=float)
    
    def date(self, code=None, freq=None):
        if code is None:
            code = self._default_stock_id
        if freq is None:
            freq = self._defaut_freq
        df = self._data_loader.get_stock_data_by_date(code, freq)
        if df.empty:
            return np.array([])
        return np.array(df['Date'], dtype=str)

    @property
    def close_default(self):
        return self.close(self._default_stock_id, self._defaut_freq)
    
    @property
    def high_default(self):
        return self.high(self._default_stock_id, self._defaut_freq)
    
    @property
    def low_default(self):
        return self.low(self._default_stock_id, self._defaut_freq)
    
    @property
    def open_default(self):
        return self.open(self._default_stock_id, self._defaut_freq)
    
    @property
    def date_default(self):
        return self.date(self._default_stock_id, self._defaut_freq)



from abc import abstractmethod
import warnings
import pandas as pd

class BaseDataLoader:
    def __init__(self, start_date: str):
        self._start_date = start_date

    def get_start_date(self):
        if isinstance(self._start_date, str):
            return self._start_date
        else:
            raise ValueError("Start date must be a string representing YYYYMMDD.")
        
    @abstractmethod
    def get_stock_data(self, code, freq):
        pass

    #检查第一个时间是否是self._start_date, 然后根据开始时间对齐获得的stock数据。
    def get_stock_data_by_date(self, code, freq):
        """
        Get stock data for a specific code and frequency starting from a given date.
        """
        df = self.get_stock_data(code, freq)
        if df.empty:
            return df
        
        first_date = pd.to_datetime(df['Date'].iloc[0])
        if first_date > pd.to_datetime(self.get_start_date()):
            warnings.warn(f"Data source does not start from {self.get_start_date()}. Actual start date: {first_date}.")
            return df
        
        return df[df['Date'] >= self.get_start_date()]
    
    



# -*- coding: utf-8 -*-
#

from .indicator.indicator import *
from .context import ExecutionContext

# 先定义全局变量
execution_context = None
CLOSE = None
OPEN = None
HIGH = None
LOW = None
DATE = None
C = None
O = None
H = None
L = None
D = None

def _is_freq_parameter(freq_param):
    """
    Check if the given parameter is a valid frequency parameter.
    """
    return isinstance(freq_param, str) and freq_param in ['30min', '60min', 'daily', 'weekly', 'monthly']

import functools

def _data_api_method(method_name, code=None, freq=None):
    api = execution_context.get_data_api()
    method = getattr(api, method_name)
    if _is_freq_parameter(code):
        return method(None, code)
    return method(code, freq)

def set_data_loader(data_loader, *, default_stock_id, default_freq='daily'):
    """
    Set the data loader for the execution context.
    """
    global execution_context, CLOSE, OPEN, HIGH, LOW, DATE, C, O, H, L, D
    execution_context = ExecutionContext(data_loader=data_loader, default_stock_id=default_stock_id, default_freq=default_freq)

    CLOSE = functools.partial(_data_api_method, 'close')
    HIGH = functools.partial(_data_api_method, 'high')
    LOW = functools.partial(_data_api_method, 'low')
    OPEN = functools.partial(_data_api_method, 'open')
    DATE = functools.partial(_data_api_method, 'date')

    C = execution_context.get_data_api().close_default
    O = execution_context.get_data_api().open_default
    H = execution_context.get_data_api().high_default
    L = execution_context.get_data_api().low_default
    D = execution_context.get_data_api().date_default


import types as _types
import sys as _sys

# 自动收集 indicator.py 中的所有函数名
_indicator_module = _sys.modules[__name__].__dict__
_indicator_funcs = [
    name for name, obj in _indicator_module.items()
    if isinstance(obj, _types.FunctionType) and obj.__module__.endswith("indicator")
]

__all__ = [
    # ...其它导出...
    "set_data_loader", "CLOSE", "OPEN", "HIGH", "LOW", "DATE", "C", "O", "H", "L", "D"
] + _indicator_funcs


from .data.data_api import DataApi

class ExecutionContext(object):
    def __init__(self, data_loader, default_stock_id, default_freq):
        self._set_default_freq(default_freq)
        self._set_default_stock_id(default_stock_id)
        self._data_api = DataApi(data_loader, self._default_stock_id, self._default_freq)

    def _set_default_stock_id(self, stock_id):
        if not isinstance(stock_id, str):
            raise ValueError("Default stock ID must be a string.")
        self._default_stock_id = stock_id

    def _set_default_freq(self, freq):
        valid_freqs = ['30min', '60min', 'daily', 'weekly', 'monthly']
        if freq not in valid_freqs:
            raise ValueError(f"Invalid frequency: {freq}. Must be one of {valid_freqs}.")
        self._default_freq = freq

    def get_data_api(self):
        return self._data_api






import codecs
import os
from setuptools import setup, find_packages

# these things are needed for the README.md show on pypi
here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()


VERSION = '0.1.5'
DESCRIPTION = 'aiwang-indicator 是一个采用通达信指标接口的 Python 技术指标编辑公式库。用户可以像在通达信中一样,使用简洁的TDX style语法进行各类技术指标的编写与回测,非常适合量化研究和策略开发。'

# Setting up
setup(
    name="aiwang-indicator",
    version=VERSION,
    author="aiwangwang",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        "numpy",
        "pandas",
        "akshare",
    ],
    keywords=["通达信指标", 'TDX', "技术指标", "量化"],
)


from .base_loader import BaseDataLoader
import pandas as pd
import akshare as ak

class AkLoader(BaseDataLoader):

    def __init__(self, start_date):
        super().__init__(start_date)

    def get_stock_data(self, code, freq):
        if freq in ['30min', '60min']:
            full_code = f'sh{code}' if code.startswith('6') else f'sz{code}'
            freq_para = '30' if freq == '30min' else '60'
            df = ak.stock_zh_a_minute(symbol=full_code, period=freq_para, adjust="qfq")
        else:
            df = ak.stock_zh_a_hist(symbol=code, period=freq, start_date=self.get_start_date(), adjust="qfq")
        return self._convert_to_standard_df(df, freq)

    def _convert_to_standard_df(self, df, freq):
        """
        Convert a fetched dataframe to a standard format with columns:
        ['Date', 'Open', 'High', 'Low', 'Close'].

        The function handles two types of input dataframes:
        1. Chinese column names: ['日期', '股票代码', '开盘', '收盘', '最高', '最低', '成交量']
        where '日期' -> 'Date', '开盘' -> 'Open', '最高' -> 'High',
        '最低' -> 'Low', '收盘' -> 'Close'.

        2. English column names: ['day', 'open', 'high', 'low', 'close', 'volume']
        where 'day' -> 'Date', 'open' -> 'Open', 'high' -> 'High',
        'low' -> 'Low', 'close' -> 'Close'.
        """
        if '日期' in df.columns:
            df = df.rename(columns={
                '日期': 'Date',
                '开盘': 'Open',
                '最高': 'High',
                '最低': 'Low',
                '收盘': 'Close'
            })
        elif 'day' in df.columns:
            df = df.rename(columns={
                'day': 'Date',
                'open': 'Open',
                'high': 'High',
                'low': 'Low',
                'close': 'Close'
            })
        else:
            raise ValueError("DataFrame schema not recognized.")
        
        # Select only the standard columns.
        df = df[['Date', 'Open', 'High', 'Low', 'Close']]
        # Convert OHLC columns to numeric types
        df[['Open', 'High', 'Low', 'Close']] = df[['Open', 'High', 'Low', 'Close']].apply(pd.to_numeric, errors='coerce')
        if freq in ['30min', '60min']:
            df['Date'] = pd.to_datetime(df['Date'], errors='coerce').dt.strftime('%Y%m%d %H:%M:%S')
        else:
            df['Date'] = pd.to_datetime(df['Date'], errors='coerce').dt.strftime('%Y%m%d')
        return df


#以下所有函数如无特别说明，输入参数S均为numpy序列或者列表list，N为整型int 
#应用层1级函数完美兼容通达信或同花顺，具体使用方法请参考通达信 

import numpy as np
import pandas as pd

#------------------ 0级：核心工具函数 --------------------------------------------      
def RD(N,D=3):   return np.round(N,D)        #四舍五入取3位小数 
def RET(S,N=1):  return np.array(S)[-N]      #返回序列倒数第N个值,默认返回最后一个
def ABS(S):      return np.abs(S)            #返回N的绝对值
def LN(S):       return np.log(S)            #求底是e的自然对数,
def POW(S,N):    return np.power(S,N)        #求S的N次方
def SQRT(S):     return np.sqrt(S)           #求S的平方根
def SIN(S):      return np.sin(S)            #求S的正弦值（弧度)
def COS(S):      return np.cos(S)            #求S的余弦值（弧度)
def TAN(S):      return np.tan(S)            #求S的正切值（弧度)  
def MAX(S1,S2):  return np.maximum(S1,S2)    #序列max
def MIN(S1,S2):  return np.minimum(S1,S2)    #序列min
def IF(S,A,B):   return np.where(S,A,B)      #序列布尔判断 return=A  if S==True  else  B


def REF(S, N=1):          #对序列整体下移动N,返回序列(shift后会产生NAN)    
    return pd.Series(S).shift(N).values  

def DIFF(S, N=1):         #前一个值减后一个值,前面会产生nan 
    return pd.Series(S).diff(N).values     #np.diff(S)直接删除nan，会少一行

def STD(S,N):             #求序列的N日标准差，返回序列    
    return  pd.Series(S).rolling(N).std(ddof=0).values     

def SUM(S, N):            #对序列求N天累计和，返回序列    N=0对序列所有依次求和         
    return pd.Series(S).rolling(N).sum().values if N>0 else pd.Series(S).cumsum().values  

def CONST(S):             #返回序列S最后的值组成常量序列
    return np.full(len(S),S[-1])
  
def HHV(S,N):             #HHV(C, 5) 最近5天收盘最高价        
    return pd.Series(S).rolling(N).max().values     

def LLV(S,N):             #LLV(C, 5) 最近5天收盘最低价     
    return pd.Series(S).rolling(N).min().values    
    
def HHVBARS(S,N):         #求N周期内S最高值到当前周期数, 返回序列
    return pd.Series(S).rolling(N).apply(lambda x: np.argmax(x[::-1]),raw=True).values 

def LLVBARS(S,N):         #求N周期内S最低值到当前周期数, 返回序列
    return pd.Series(S).rolling(N).apply(lambda x: np.argmin(x[::-1]),raw=True).values    
  
def MA(S,N):              #求序列的N日简单移动平均值，返回序列                    
    return pd.Series(S).rolling(N).mean().values  
  
def EMA(S,N):             #指数移动平均,为了精度 S>4*N  EMA至少需要120周期     alpha=2/(span+1)    
    return pd.Series(S).ewm(span=N, adjust=False).mean().values     

def SMA(S, N, M=1):       #中国式的SMA,至少需要120周期才精确 (雪球180周期)    alpha=1/(1+com)    
    return pd.Series(S).ewm(alpha=M/N,adjust=False).mean().values           #com=N-M/M

def WMA(S, N):            #通达信S序列的N日加权移动平均 Yn = (1*X1+2*X2+3*X3+...+n*Xn)/(1+2+3+...+Xn)
    return pd.Series(S).rolling(N).apply(lambda x:x[::-1].cumsum().sum()*2/N/(N+1),raw=True).values 

def DMA(S, A):            #求S的动态移动平均，A作平滑因子,必须 0<A<1  (此为核心函数，非指标）
    if isinstance(A,(int,float)):  return pd.Series(S).ewm(alpha=A,adjust=False).mean().values    
    A=np.array(A);   A[np.isnan(A)]=1.0;   Y= np.zeros(len(S));   Y[0]=S[0]     
    for i in range(1,len(S)): Y[i]=A[i]*S[i]+(1-A[i])*Y[i-1]      #A支持序列 by jqz1226         
    return Y             
  
def AVEDEV(S, N):         #平均绝对偏差  (序列与其平均值的绝对差的平均值)   
    return pd.Series(S).rolling(N).apply(lambda x: (np.abs(x - x.mean())).mean()).values 

def SLOPE(S, N):          #返S序列N周期回线性回归斜率            
    return pd.Series(S).rolling(N).apply(lambda x: np.polyfit(range(N),x,deg=1)[0],raw=True).values

def FORCAST(S, N):        #返回S序列N周期回线性回归后的预测值， jqz1226改进成序列出    
    return pd.Series(S).rolling(N).apply(lambda x:np.polyval(np.polyfit(range(N),x,deg=1),N-1),raw=True).values  

def LAST(S, A, B):        #从前A日到前B日一直满足S_BOOL条件, 要求A>B & A>0 & B>=0 
    return np.array(pd.Series(S).rolling(A+1).apply(lambda x:np.all(x[::-1][B:]),raw=True),dtype=bool)

def INDEX(S):             #将序列S转化为指数化序列，首日为1000，后续按涨跌幅递推
    S = np.array(S, dtype=float)
    idx = np.zeros(len(S))
    if len(S) == 0:
        return idx
    idx[0] = 1000
    for i in range(1, len(S)):
        if S[i-1] == 0:
            idx[i] = idx[i-1]
        else:
            idx[i] = idx[i-1] * (S[i] / S[i-1])
    return idx

def CUMPROD(S):              #对输入序列S进行累计乘积，返回同长度的序列。例如：S=[1,2,3,4]，返回[1,2,6,24]
    S = np.array(S, dtype=float)
    # 将nan替换为1（累计乘积的中性元素）
    S[np.isnan(S)] = 1
    return np.cumprod(S)
  
#------------------   1级：应用层函数(通过0级核心函数实现）使用方法请参考通达信--------------------------------
def COUNT(S, N):                       # COUNT(CLOSE>O, N):  最近N天满足S_BOO的天数  True的天数
    return SUM(S,N)    

def EVERY(S, N):                       # EVERY(CLOSE>O, 5)   最近N天是否都是True
    return  IF(SUM(S,N)==N,True,False)                    
  
def EXIST(S, N):                       # EXIST(CLOSE>3010, N=5)  n日内是否存在一天大于3000点  
    return IF(SUM(S,N)>0,True,False)

def FILTER(S, N):                      # FILTER函数，S满足条件后，将其后N周期内的数据置为0, FILTER(C==H,5)
    for i in range(len(S)): S[i+1:i+1+N]=0  if S[i] else S[i+1:i+1+N]        
    return S                           # 例：FILTER(C==H,5) 涨停后，后5天不再发出信号 
  
def BARSLAST(S):                       #上一次条件成立到当前的周期, BARSLAST(C/REF(C,1)>=1.1) 上一次涨停到今天的天数 
    M=np.concatenate(([0],np.where(S,1,0)))  
    for i in range(1, len(M)):  M[i]=0 if M[i] else M[i-1]+1    
    return M[1:]                       

def BARSLASTCOUNT(S):                  # 统计连续满足S条件的周期数        by jqz1226
    rt = np.zeros(len(S)+1)            # BARSLASTCOUNT(CLOSE>OPEN)表示统计连续收阳的周期数
    for i in range(len(S)): rt[i+1]=rt[i]+1  if S[i] else rt[i+1]
    return rt[1:]  
  
def BARSSINCEN(S, N):                  # N周期内第一次S条件成立到现在的周期数,N为常量  by jqz1226
    return pd.Series(S).rolling(N).apply(lambda x:N-1-np.argmax(x) if np.argmax(x) or x[0] else 0,raw=True).fillna(0).values.astype(int)
  
def CROSS(S1, S2):                     # 判断向上金叉穿越 CROSS(MA(C,5),MA(C,10))  判断向下死叉穿越 CROSS(MA(C,10),MA(C,5))   
    return np.concatenate(([False], np.logical_not((S1>S2)[:-1]) & (S1>S2)[1:]))    # 不使用0级函数,移植方便  by jqz1226
    
def LONGCROSS(S1,S2,N):                # 两条线维持一定周期后交叉,S1在N周期内都小于S2,本周期从S1下方向上穿过S2时返回1,否则返回0         
    return  np.array(np.logical_and(LAST(S1<S2,N,1),(S1>S2)),dtype=bool)            # N=1时等同于CROSS(S1, S2)
    
def VALUEWHEN(S, X):                   # 当S条件成立时,取X的当前值,否则取VALUEWHEN的上个成立时的X值   by jqz1226
    return pd.Series(np.where(S,X,np.nan)).ffill().values  

def BETWEEN(S, A, B):                  # S处于A和B之间时为真。 包括 A<S<B 或 A>S>B
    return ((A<S) & (S<B)) | ((A>S) & (S>B))  

def TOPRANGE(S):                       # TOPRANGE(HIGH)表示当前最高价是近多少周期内最高价的最大值 by jqz1226
    rt = np.zeros(len(S))
    for i in range(1,len(S)):  rt[i] = np.argmin(np.flipud(S[:i]<S[i]))
    return rt.astype('int')

def LOWRANGE(S):                       # LOWRANGE(LOW)表示当前最低价是近多少周期内最低价的最小值 by jqz1226
    rt = np.zeros(len(S))
    for i in range(1,len(S)):  rt[i] = np.argmin(np.flipud(S[:i]>S[i]))
    return rt.astype('int')
